#!/usr/bin/python
# -*- coding: utf-8 -*-
##********************************************************************************************************************************************************
##
##  This module converts a fits file to an ASCII file using pyfits
##  Copyright (C) 2009 - 2016  Thomas Moeller
##
##  I. Physikalisches Institut, University of Cologne
##
##
##
##  Versions of the program:
##
##  Who           When         What
##
##  T. Moeller    2013-02-14   initial version
##  T. Moeller    2016-11-22   include cos correction
##
##
##
##  License:
##
##    GNU GENERAL PUBLIC LICENSE
##    Version 3, 29 June 2007
##    (Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>)
##
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##********************************************************************************************************************************************************


##******************************************************************** load packages *********************************************************************
import numpy                                                                                ## import numpy package
import os.path                                                                              ## load package for file management
import os                                                                                   ## import os package
import string                                                                               ## load string package
import shutil                                                                               ## import shutil package
import sys                                                                                  ## import sys package
from math import *                                                                          ## import math package


## import pyfits package via astropy
try:
    import astropy.io.fits as pyfits
except ImportError:
    import pyfits
##--------------------------------------------------------------------------------------------------------------------------------------------------------


##--------------------------------------------------------------------------------------------------------------------------------------------------------
##
## start main program
##
if __name__ == '__main__':


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## get OnlyYColumn
    OnlyYColumn = sys.argv[1]


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## set print flag
    printflag = "false"


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## check if FitsFilename ends with .fits
    FitsFilename = sys.argv[2]                                                              ## get fits file name
    if not FitsFilename.endswith(".fits"):
        print "\n\t Error in fits2dat subroutine!"
        print "\t\t The selected file ",FitsFilename
        print "\t\t does not end with '.fits'!"
        print "\t\t Please correct input!\n"
        sys.exit(0)


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## check, if filename exists
    if (os.path.exists(FitsFilename) and os.path.isfile(FitsFilename)):


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## open fits-file using pyfits
        hdulist = pyfits.open(FitsFilename)                                                 ## open fits file
        fitsdata = hdulist[0].data                                                          ## read fits data
        lengthFITSDATA = fitsdata.shape                                                     ## determine length in each dimension
        NumDimensions = len(lengthFITSDATA)                                                 ## determine dimension of fits file
        hdulist.close()                                                                     ## close fits-file

        # Debug:
        #    print "Number of columns = ",NumDimensions
        #    print "lengthFITSDATA = ",lengthFITSDATA


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## open dat-file name
        CurrentPath = os.getcwd() + "/"
        DatFilename = FitsFilename.replace(".fits", ".dat")                                 ## define name of output ASCII file
        DatFilename = DatFilename.strip()
        if (DatFilename == ""):
            print "\n\n\n\t Error in subroutine fits2dat.py:"
            print "\t\t The given file name is empty. DatFilename = ", DatFilename
            print "\n\n"
            sys.exit(0)
        else:
            if (DatFilename[0] == "/"):
                DatFile = open(DatFilename, "w")                                            ## open dat-file name without extended file name
            else:
                DatFile = open(CurrentPath + DatFilename, "w")                              ## open dat-file name with extended file name


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## read header information


        ## get number of axes
        try:
            NAXIS = pyfits.getval(FitsFilename, 'NAXIS', 0)
        except KeyError, err:
            NAXIS = 1

        # Debug:
        # print 'NAXIS = ',NAXIS


        ## get scale parameter BSCALE
        try:
            BSCALE = pyfits.getval(FitsFilename, 'BSCALE', 0)
        except KeyError, err:
            BSCALE = 1

        # Debug:
        # print '\n\nBSCALE = ',BSCALE


        ## get zero parameter BZERO
        try:
            BZERO = pyfits.getval(FitsFilename, 'BZERO', 0)
        except KeyError, err:
            BZERO = 0.0

        # Debug:
        # print '\n\nBZERO = ',BZERO


        ## read further header information for each axis
        CTYPEList = []
        CRPIXList = []
        CRVALList = []
        CDELTList = []
        length = []
        ValueOfFirstPointList = []
        NumDimensions = 0
        for i in xrange(NAXIS):                                                             ## loop over all axis
            try:
                lenAxis = pyfits.getval(FitsFilename, 'NAXIS' + str(i + 1), 0)              ## read NAXIS
            except KeyError, err:
                lenAxis = 1
            if (lenAxis > 0):                                                               ## consider only those axis with at least one data point
                length.append(lenAxis)
                NumDimensions += 1
                try:
                    CTYPEValue = pyfits.getval(FitsFilename, 'CTYPE' + str(i + 1), 0)       ## read CTYPE
                except KeyError, err:
                    CTYPEValue = ""
                CTYPEList.append(CTYPEValue)

                try:
                    CRPIXValue = pyfits.getval(FitsFilename, 'CRPIX' + str(i + 1), 0)       ## read CRPIX
                except KeyError, err:
                    CRPIXValue = 1
                CRPIXList.append(CRPIXValue)

                try:
                    CRVALValue = pyfits.getval(FitsFilename, 'CRVAL' + str(i + 1), 0)       ## read CRVAL
                except KeyError, err:
                    CRVALValue = 1
                CRVALList.append(CRVALValue)

                try:
                    CDELTValue = pyfits.getval(FitsFilename, 'CDELT' + str(i + 1), 0)       ## read CDELT
                except KeyError, err:
                    CDELTValue = 1
                CDELTList.append(CDELTValue)


                ## calculate value at first pixel
                ValueOfFirstPoint = CRVALValue - ((CRPIXValue - 1) * CDELTValue)
                ValueOfFirstPointList.append(ValueOfFirstPoint)


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## add cos correction for R.A., if FITS data describe astronomical data (map/cube)
        for AxisID, LocalCTYPE in enumerate(CTYPEList):
            LocalCTYPE = LocalCTYPE.strip()
            LocalCTYPE = LocalCTYPE.lower()
            if (LocalCTYPE.startswith("ra--") and CRPIXList[AxisID] != 1):
                CRVAL = CRVALList[AxisID]
                CRPIX = CRPIXList[AxisID]
                CDELT = CDELTList[AxisID]
                DECRefVal = 0.0
                for AxisIDDEC, LocalCTYPEDEC in enumerate(CTYPEList):
                    LocalCTYPEDEC = LocalCTYPEDEC.strip()
                    LocalCTYPEDEC = LocalCTYPEDEC.lower()
                    if (LocalCTYPEDEC.startswith("dec-")):
                        DECRefVal = CRVALList[AxisIDDEC]
                        break
                CDELT = CDELT / cos(pi / 180.0 * DECRefVal)                                 ## add cos correction to delta of R.A.
                CDELTList[AxisID] = CDELT
                ValueOfFirstPoint = CRVAL - ((CRPIX - 1) * CDELT)                           ## calculate new (corrected) first point
                ValueOfFirstPointList[AxisID] = ValueOfFirstPoint

        # Debug:
        # print '\n\nlength = ',length
        # print 'CTYPEList = ',CTYPEList
        # print 'CRPIXList = ',CRPIXList
        # print 'CRVALList = ',CRVALList
        # print 'CDELTList = ',CDELTList
        # print "lenAxis = ", lenAxis
        # print "ValueOfFirstPointList = ", ValueOfFirstPointList


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## check, if data cube has to be transposed
        if (lengthFITSDATA != length and NumDimensions > 1):
            fitsdata = numpy.transpose(fitsdata)

            # Debug:
            # print "fitsdata.shape =  ", fitsdata.shape


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## write data to dat file depending on the dimension of the fits file
        if (NumDimensions == 1):
            for x in xrange(length[0]):
                xPixel = (ValueOfFirstPointList[0] + (x * CDELTList[0])) * BSCALE + BZERO
                if (OnlyYColumn == "y"):
                    line = "%25.16e" % fitsdata[x]
                else:
                    line = "%25.16e %25.16e" % (xPixel, fitsdata[x])
                DatFile.write(line + "\n")


        elif (NumDimensions == 2):
            for x in xrange(length[0]):
                xPixel = (ValueOfFirstPointList[0] + (x * CDELTList[0])) * BSCALE + BZERO
                for y in xrange(length[1]):
                    yPixel = (ValueOfFirstPointList[1] + (y * CDELTList[1])) * BSCALE + BZERO
                    if (OnlyYColumn == "y"):
                        line = "%25.16e" % fitsdata[x, y]
                    else:
                        line = "%25.16e %25.16e %25.16e" % (xPixel, yPixel, fitsdata[x, y])
                    DatFile.write(line + "\n")


        elif (NumDimensions == 3):
            for x in xrange(length[0]):
                xPixel = (ValueOfFirstPointList[0] + (x * CDELTList[0])) * BSCALE + BZERO
                for y in xrange(length[1]):
                    yPixel = (ValueOfFirstPointList[1] + (y * CDELTList[1])) * BSCALE + BZERO
                    for z in xrange(length[2]):
                        zPixel = (ValueOfFirstPointList[2] + (z * CDELTList[2])) * BSCALE + BZERO
                        if (OnlyYColumn == "y"):
                            line = "%25.16e" % fitsdata[x, y, z]
                        else:
                            line = "%25.16e %25.16e %25.16e %25.16e" % (xPixel, yPixel, zPixel, fitsdata[x, y, z])
                        DatFile.write(line + "\n")


        elif (NumDimensions == 4):
            for x in xrange(length[0]):
                xPixel = (ValueOfFirstPointList[0] + (x * CDELTList[0])) * BSCALE + BZERO
                for y in xrange(length[1]):
                    yPixel = (ValueOfFirstPointList[1] + (y * CDELTList[1])) * BSCALE + BZERO
                    for z in xrange(length[2]):
                        zPixel = (ValueOfFirstPointList[2] + (z * CDELTList[2])) * BSCALE + BZERO
                        for t in xrange(length[3]):
                            tPixel = (ValueOfFirstPointList[3] + (t * CDELTList[3])) * BSCALE + BZERO
                            if (OnlyYColumn == "y"):
                                line = "%25.16e" % fitsdata[x, y, z, t]
                            else:
                                line = "%25.16e %25.16e %25.16e %25.16e %25.16e" % (xPixel, yPixel, zPixel, tPixel, fitsdata[x, y, z, t])
                            DatFile.write(line + "\n")


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## close dat file
    DatFile.close()


## finished!
##--------------------------------------------------------------------------------------------------------------------------------------------------------
##--------------------------------------------------------------------------------------------------------------------------------------------------------
##--------------------------------------------------------------------------------------------------------------------------------------------------------

